# -*- encoding:utf-8 -*-
"""
@author：lin.li
@filename：__init__.py
@time：2022/8/15  21:36
@file description: pubmed_download
"""
from .pubmed_download import DownloadPdf
from .pubmed_mapper import Article
